# This script generates a list of OpenGL functions and their corresponding
# documentation page on www.opengl.org in file GLSpec/glRef.txt.
# This list is generated by checking the function names against existing
# reference and extension specification pages with the http package.
# As this process takes some time, it must be executed by hand in a 2-step
# bootstrap process, once a new GLEW version is incorporated into Tcl3D.
# 
# See the ReadmeGlew.txt file in directory GLSpec for a detailed description
# of the update and bootstrap process.

package require http
package require tcl3d

# File where the list of function names and corresponding URL's is stored.
set outFile "GLSpec/glRef.txt"

# Template for standard OpenGL and GLU function reference pages.
# Note, that currently only reference pages exist up to version 2.1.
set glUrlTmpl  "http://www.opengl.org/sdk/docs/man/xhtml/%s.xml"

# Template for GLUT function reference pages.
# This is not a real template, but the master page of the GLUT API reference.
set glutUrlTmpl "http://www.opengl.org/documentation/specs/glut/spec3/spec3.html"

# Template for OpenGL extension function specification pages.
set extUrlTmpl "http://www.opengl.org/registry/specs/%s/%s.txt"

# Template for OpenGL functions, where no URL has been found.
# Search on Google for the function name.
set naUrlTmpl  "http://www.google.com/search?q=%s"

# Procedure checking the existence of a specified URL.
# The check is done by comparing the returned http string against "404".
proc UrlExists { url } {
    global gUrlChecked

    set retVal true
    if { [info exists gUrlChecked($url)] } {
        # puts "       Using cache $gUrlChecked($url) for $url"
        set retVal $gUrlChecked($url)
    } else {
        set token [http::geturl $url -binary 1]
        upvar #0 $token state
        if { [string match "*404*" $state(http)] } {
            set retVal false
        }
        ::http::cleanup $token
        set gUrlChecked($url) $retVal
    }
    return $retVal
}

# If the reference file already exists, create a backup copy.
if { [file exists $outFile] } {
    file copy $outFile $outFile.bak
}

set outFp [open $outFile "w"]

set glAllFuncList [tcl3dOglGetFuncList]
set numFuncs [llength $glAllFuncList]

set minLength 10000
foreach func $glAllFuncList {
    set len [string length $func]
    if { $len < $minLength } {
        set minLength $len
    }
}
puts "Shortest function name has $minLength characters"
incr minLength -1

# Change the start parameter only for testing purposes.
set start 0
# set start 550 ; # Start of GL3 functions
# set start 610 ; # Start of extension functions

set glInd 0
foreach func $glAllFuncList {
    if { $glInd < $start } {
        incr glInd
        continue
    }
    set glVersion [tcl3dOglGetFuncVersion $func]
    set foundUrl false
    set isGL3    false
    if { ! [string match "GL_VERSION*" $glVersion] && \
         $glVersion ne "GL_ARB_imaging" } {
        # The function is an extension function. Create links
        # to the OpenGL extension repositiory.
        set extList [split $glVersion "_"]
        set extSuffix [join [lrange $extList 2 end] "_"]
        if { [lindex $extList 1] eq "OES" } {
            # OES extension URL's are named a little bit different.
            set extSuffix [format "OES_%s" $extSuffix]
        }
        set url [format $extUrlTmpl [lindex $extList 1] $extSuffix]
        if { [UrlExists $url] } {
            puts $outFp "$func $url"
            puts [format "%4d of %4d: %s %s" $glInd $numFuncs $func $url]
            set foundUrl true
        }
        # Some extensions (ex. GL_EXT_multisample) started out as a WGL_*
        # extension, and are therefore available under that URL.
        set url [format $extUrlTmpl [lindex $extList 1] "wgl_$extSuffix"]
        if { [UrlExists $url] } {
            puts $outFp "$func $url"
            puts [format "%4d of %4d: %s %s" $glInd $numFuncs $func $url]
            set foundUrl true
        }
    } elseif { [string match "GL_VERSION_3_*" $glVersion] } {
        # The function is an OpenGL 3.X function. No reference pages
        # exist currently. Create links to search with Google.
        set isGL3 true
    } else {
        # The function is an OpenGL standard function or a function from
        # the GL_ARB_imaging extension. Find the corresponding
        # reference page. This is done by removing characters from the end of
        # the function name until a valid reference page URL is found.
        # This is needed, because there is only one page for a class of 
        # functions, ex. page glMap.xml for glFogi, glFogiv, glFogf, glFogfv.
        set baseFunc $func
        set len [string length $func]
        set ind [expr $len -1]
        while { ! $foundUrl && $ind >= $minLength } {
            set baseFunc [string range $func 0 $ind]
            set url [format $glUrlTmpl $baseFunc]
            if { [UrlExists $url] } {
                puts $outFp "$func $url"
                puts [format "%4d of %4d: %s %s" $glInd $numFuncs $func $url]
                set foundUrl true
            }
            incr ind -1
        }
    }
    if { ! $foundUrl } {
        set naUrl [format $naUrlTmpl $func]
        if { $isGL3 } {
            append naUrl "+$glVersion"
        }
        puts $outFp "$func $naUrl"
        puts [format "%4d of %4d: %s %s" $glInd $numFuncs $func $naUrl]
    }
    flush $outFp
    flush stdout
    incr glInd
}

# Now search for all GLU and GLUT functions.
set gluInd 0
set gluAllFuncList [info commands glu*]
set numFuncs [llength $gluAllFuncList]
foreach func $gluAllFuncList {
    if { [string match "glut*" $func] } {
        set url $glutUrlTmpl
        puts $outFp "$func $url"
        puts [format "%4d of %4d: %s %s" $gluInd $numFuncs $func $url]
    } else {
        set url [format $glUrlTmpl $func]
        if { [UrlExists $url] } {
            puts $outFp "$func $url"
            puts [format "%4d of %4d: %s %s" $gluInd $numFuncs $func $url]
        } else {
            set naUrl [format $naUrlTmpl $func]
            puts $outFp "$func $naUrl"
            puts [format "%4d of %4d: %s %s" $gluInd $numFuncs $func $naUrl]
        }
    }
    flush $outFp
    flush stdout
    incr gluInd
}

close $outFp
exit 0
